// Demonstrates the use of the GLUT - and the CCD IK algorithm
//
// NOTE: BEFORE LOOKING AT THE CODE - MAKE SURE YOU UNDERSTAND THE OPERATION
//       OF THE PROGRAM AND HAVE READ THE ACCOMPANYING NOTES

// All matrix operation tke use the conventional transformation notation
// i.e.  the transformation PRE-multiplies the position vector.  p' = [a]p
// 

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>

#ifdef MACOSX
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
#else
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

// MACROS for Vector arithmetic
#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECCOPY4(a,b)   { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; b[3] = a[3];}
#define INVERT(a)       { a[0] = -a[0]; a[1] = -a[1]; a[2] = -a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB4(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2]; c[3]=a[3]-b[3];}
#define VECSUM4(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2]; c[3]=a[3]+b[3];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define VECSCALE4(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2]; c[3]=b[3];}
#define VECSUMSCALE(s,a,b,c)  \
                        { c[0]=s*(a[0]+b[0]); c[1]=s*(a[1]+b[1]); c[2]=s*(a[2]+b[2]);}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define DOT4(a,b)       ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                      }
#define CROSS4(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                          r[3] = 1.0; \
                      }
// Other macros
#define DEGRAD 3.1415926/180.0
#define RADDEG 180.0/3.1415926

// ROBOT dimensions
#define BASEHEIGHT  31.5
#define UPPERARM    45.0
#define LOWERARM    50.0
#define HAND         8.0
#define TOFFSET     15.0

// These are in the ROBOT coord system  X to front, Y to side, Z up  (dimensions cms)
#define BASE_X -55    
#define BASE_Y   0
#define BASE_Z  45
#define SIZE_X   8
#define SIZE_Y   8
#define SIZE_Z   8

typedef float  vector[3];

// These angles control the angles of rotation away from the 
// rest position for the Robot Arm's joints, they are as indicated
// in the accompanying document.
GLfloat rs1=0.0,rs2=0.0,rs3=0.0,
        re1=0.0,re2=0.0,
        rw1=0.0,rw2=0.0;

// These angles are determined by the CCD algorithm to allow the 
//robot to move the centre of its tool
// to the given target position.  (How to arrange that the 
// arm moves its working tool so that it approached
// a working points - is a more compliacted problem. Ways to 
// solve this problem is stil an ongoing research topic.
GLfloat ccd_rs1=0.0,ccd_rs2=0.0,ccd_rs3=0.0,
        ccd_re1=0.0,ccd_re2=0.0,
        ccd_rw1=0.0,ccd_rw2=0.0;

// IK target for the robot's wrist point - when using IK mode
GLfloat p_target[3]={42,0,50};     
// These three angles are used to give the direction that the end-effector (the tool) 
// at the end of the Robot Arm should be pointing in. (When the CCD algorithm is in use.)
GLfloat hand_pan=0.0,hand_tilt=0.0,hand_twist=0.0;

// Direction and distance of the camera from the object. 
//(Used to set up the viewing transformation )
static GLdouble er=200.0,theta=0.0,phi=0.0;

// Specify the dimensions of the model robot
GLfloat baseHeight=BASEHEIGHT
       ,upperArm=UPPERARM
       ,lowerArm=LOWERARM
       ,Hand=HAND;

// User flags
int     viewdir=1       // Which view of the model  e.g. 1 -> Z axis up 
       ,modelVis=1      // Draw the model (if 0 then only draw the axes
       ,axesVis=1       // draw the axes of the frames of reference attached to each arm of the robot
       ,ccdVis=0        // draw the frames of reference calculated by the IK algorithm
       ,get_transform=0 // used to get the current transformation information and print it (so that it can be verified manually)
       ,ccdRLT=1        // display each iteration of the CCD algorithm when it is being used in ONE-SHOT mode
       ,interactive=0;  // move the target and track it interactively with the robot using IK

// transformations as determined by the OpenGL matrices
// All are (in C element order and  p = Mp transformation type)
// These transformations are all obtained by
// reading them from the OpenGL transform stack.
GLfloat m_camera[4][4],  // Camera transformation matrix 
        m_endpoint[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},  // endpoint 
        m_wristpoint[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},// wrist
        m_lowerarm[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},  
        m_upperarm[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},  
        p_upperarm[3],   // coordinate of point of upper arm
        p_lowerarm[3],   // coodinate of point of lowe arm
        p_wrist[3];      // coordinate of ROBOT wrist

// This group of variables are used in the CCD IK algorithm - they mimic the 
// forward kinematics variables used to position and draw 
// the robot. However the robot is drawn using OpenGL transformation. 
// The IK algorithm needs to use its own transformations
// while it is calculating the correct joint angles.
float   c_upperarm[4]={0,0,BASEHEIGHT,1},            
        c_lowerarm[4]={0,0,BASEHEIGHT+UPPERARM,1},
        c_wrist[4]={0,0,BASEHEIGHT+UPPERARM+LOWERARM,1},
        c_hand[4]={0,0,BASEHEIGHT+UPPERARM+LOWERARM+HAND,1},
        // this is a point used for the hand to point at, in the algorithm
        c_target[4]={TOFFSET,0,BASEHEIGHT+UPPERARM+LOWERARM,1},  
        II[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},   // Identity matrix
        B[4][4],  // base transform
        U[4][4],  // upper arm transform
        L[4][4],  // lower arm transform
        H[4][4],  // hand transform
        // composites
        BU[4][4],  
        BUL[4][4],
        BULH[4][4],
        // matrix giving orienation of the tool 
        ToolO[4][4]={1,0,0,0, 0,1,0,0,  0,0,1,0,  0,0,0,1};  


// function prototypes
static void InvertMatrixRotation(float T[4][4], float t[4][4], int zero_translation);
static void InvertMatrixRotationT(float T[4][4], float t[4][4], int not_used);
static void TransposeMatrix16(GLfloat m[16]);
static void m4by4(GLfloat t1[4][4], GLfloat t2[4][4], GLfloat tr[4][4]);
static void m16to4(GLfloat m[16], GLfloat v[4][4]);
static void m4to16(GLfloat v[4][4], GLfloat m[16]);
static void m4by1(GLfloat t[4][4], GLfloat m[4],  GLfloat r[4]);
static void rotz(float tr[4][4], float ang);
static void rotx(float tr[4][4], float ang);
static void roty(float tr[4][4], float ang);
static void tram(float t[4][4], float dx, float dy, float dz);
static void c4to4(float tin[4][4], float tout[4][4]);
static void null_transform(float t[4][4]);
static void ExtractMatrixRotation(float T[4][4], float t[4][4]);
static void SetMatrixRotation(float T[4][4], float r[4][4], float t[4][4]);
static void RunCCD(int prnt);
static void normalise(float v[4]);
static void RotateVector(float ct, float st, float r[4], float n[4]);
static void SetHandDirection();
static void display(void);

/////////////////////////////////CCD IK algorithm functions //////////////////

static void CalculateCcdTransforms(void){
 // This function calcuates the transformations that are to be applied to the arms of the robot based on
 // the current values of pivots applied to the axes of the robot, and as determined by CCD IK algorithm
 float zero[4]={0,0,0,1};
 float m1[4][4];
 float s1[4][4],s2[4][4],s3[4][4],e1[4][4],e2[4][4],w1[4][4],w2[4][4],tb[4][4],tua[4][4],tla[4][4],th[4][4];
 rotz(s1,ccd_rs1);
 roty(s2,ccd_rs2);
 tram(tb,0.0,0.0,baseHeight);
 m4by4(tb,s2,m1);
 m4by4(s1,m1,B);  // base transformation
 rotz(s3,ccd_rs3);
 roty(e1,ccd_re1);
 tram(tua,0.0,0.0,upperArm);
 m4by4(tua,e1,m1);
 m4by4(s3,m1,U);  // upperarm transform - relative to base
 rotz(e2,ccd_re2);
 roty(w1,ccd_rw1);
 tram(tla,0.0,0.0,lowerArm);
 m4by4(tla,w1,m1);
 m4by4(e2,m1,L); // lowerarm - relaive to upperarm
 rotz(w2,ccd_rw2);
 tram(th,0.0,0.0,Hand);
 m4by4(w2,th,H); // Hand relative to lowerarm 
 m4by4(B,U,BU);   // composites
 m4by4(BU,L,BUL);  
 m4by4(BUL,H,BULH);
 // now apply the transformation to the positions of the links
 m4by1(B,zero,c_upperarm);  //at top of base
 m4by1(BU,zero,c_lowerarm);
 m4by1(BUL,zero,c_wrist);   // this is the primary pivot point
 m4by1(BULH,zero,c_hand);   // end point
}

static int TargetReached(float *e, float p_t[]){
 float d,errorD=0.1,dend[4];
 VECSUB4(p_t,c_wrist,dend)
 d=sqrt(dend[0]*dend[0]+dend[1]*dend[1]+dend[2]*dend[2]);
 *e=d;
 if(d < errorD)return 1;
 return 0;
}

static void SetHandDirection(void){ // Used only with CCD algorithm to keep the tool "Hand" pointing in the same direction
 float d;
 float ddv1[4],ddv2[4],up[4]={0,0,1,1},dirn[4]={1,0,0,1};
 float mmr[4][4],mmi[4][4],hh[4][4];
 int i,j;

 float r1[4][4],r2[4][4],r3[4][4],ti[4][4];
 dirn[0]=cos(hand_pan*DEGRAD)*cos(hand_tilt*DEGRAD);
 dirn[1]=sin(hand_pan*DEGRAD)*cos(hand_tilt*DEGRAD);
 dirn[2]=sin(hand_tilt*DEGRAD);
 normalise(dirn);
 rotz(r1,hand_twist);         // make transform for end of tool effector
 roty(r2,90.0f-hand_tilt);   // to send to robot
 rotz(r3,hand_pan);
 m4by4(r2,r1,ti);
 m4by4(r3,ti,ToolO);

 VECSUM4(c_wrist,TOFFSET*dirn,c_target)  // indicate the direction "dirn" from wrist
 CalculateCcdTransforms();
 ExtractMatrixRotation(BU,mmr);   
 InvertMatrixRotation(mmr,mmi,1);
 m4by1(mmi,dirn,ddv2);                    // direction is now relative to GLOBAL axes
 ccd_re2=atan2(ddv2[1],ddv2[0])*RADDEG;
 d=sqrt(ddv2[0]*ddv2[0]+ddv2[1]*ddv2[1]);
 ccd_rw1=90.0-atan2(ddv2[2],d)*RADDEG;

 CalculateCcdTransforms();
 ExtractMatrixRotation(BUL,mmr);          // turn the hand so that its "X" axis is vertical
 InvertMatrixRotation(mmr,mmi,1); 
 m4by1(mmi,up,ddv2);
 ccd_rw2=atan2(ddv2[1],ddv2[0])*RADDEG + hand_twist;

 CalculateCcdTransforms();
 re2=ccd_re2, rw1=ccd_rw1;  rw2=ccd_rw2;  // Copy the angles to the Robit's joint angles! (the ones used to draw the robot.)
 //VECCOPY(c_hand,p_hand)   // end point
}

static void RunCCD(int prnt){ // Execute the Cyclic Coordinate descent algorithm for a TWO link arm
 int i,nsteps,stepMax=200;
 float d,ct,st,error;
 float nr[4],ddv1[4],ddv2[4];
 float mmr[4][4],mmi[4][4];
 float p_tar[3];

 // starting angles for CCD
 ccd_rs1=0.0; ccd_rs2=0.0; // upper arm angles
 ccd_rs3=0.0; ccd_re1=0.0; // lower arm angles
 ccd_re2=0.0; ccd_rw1=0.0; ccd_rw2=0.0;  // wrist hand angles (determined later)
 nsteps=0;
 for(i=0;i<3;i++)p_tar[i]=p_target[i];

 LOOP:   // Simple 2 stage CCD 
 CalculateCcdTransforms();
 if(ccdRLT && prnt){ // to display each step of the algorithm
    rs1=ccd_rs1; rs2=ccd_rs2;  rs3=ccd_rs3; re1=ccd_re1;
    display();
    Sleep(250);
 } 
 if(TargetReached(&error,p_tar))goto SOLVED;
 if(nsteps > stepMax)goto FAILED;
 // pivot lower arm - to align lower arm in direction of target 
 // (and place c_wrist on line joining c_lowerarm with p_tar)
 VECSUB4(p_tar,c_lowerarm,ddv1)
 ExtractMatrixRotation(B,mmr);   
 InvertMatrixRotationT(mmr,mmi,1);  // use Transpose assumption 
 m4by1(mmi,ddv1,ddv2);   
 ccd_rs3=atan2(ddv2[1],ddv2[0])*RADDEG;
 d=sqrt(ddv2[0]*ddv2[0]+ddv2[1]*ddv2[1]);
 ccd_re1=90.0-atan2(ddv2[2],d)*RADDEG;
 CalculateCcdTransforms();
 if(TargetReached(&error,p_tar))goto SOLVED;
 // pivot upper arm - to place c_wrist on a line joining c_upperarm with p_tar
 // this step would need to be repeated if there were more than two links in the chain
 VECSUB(c_wrist ,c_upperarm,ddv1)  normalise(ddv1);
 VECSUB(p_tar,   c_upperarm,ddv2)  normalise(ddv2);
 if(fabs(DOT4(ddv1,ddv2)) > 0.0001){ // not on line so need to rotate to it
   CROSS4(ddv1,ddv2,nr) normalise(nr);
   ct=DOT4(ddv1,ddv2); st=sqrt(1-ct*ct);
   VECSUB4(c_lowerarm,c_upperarm,ddv1)
   RotateVector(ct,st,ddv1,nr);
   ExtractMatrixRotation(II,mmr);  // This is wasteful since CCD has only two links
   InvertMatrixRotation(mmr,mmi,1); 
   m4by1(mmi,ddv1,ddv2);   
   ccd_rs1=atan2(ddv2[1],ddv2[0])*RADDEG;
   d=sqrt(ddv2[0]*ddv2[0]+ddv2[1]*ddv2[1]);
   ccd_rs2=90.0-atan2(ddv2[2],d)*RADDEG;
 } 
 nsteps++;
 //printf("Error %7.3f\n",error);
 goto LOOP; 
 FAILED: ;  // this most likely occurs when target it out of range
 SOLVED: 
 // End of CCD algorithm 
 
 // Assing the joint angles calculated by the CCD algorithm to position the arm )so that it can
 // touch the target) to the Robot
 rs1=ccd_rs1; rs2=ccd_rs2;     
 rs3=ccd_rs3; re1=ccd_re1;
 // The final joint angles will be calculated to keep the robot's hand pointing in the direction "dirn" 
 re2=0.0; rw1=0.0; rw2=0.0;  
 // now calculate the direction of the hand to keep it pointing in the desired direction
 SetHandDirection();

 if(prnt){
  printf("CCD Algorithm - Solved in %d iterations (Error %7.3f)\n",nsteps,error);
  printf("TARGET:(xyz)         (%6.2f  %6.2f %6.2f)\n", p_tar[0],p_tar[1],p_tar[2]);
  printf("CCD Positions\n");
  printf("Hand      base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_hand[0],c_hand[1],c_hand[2]);
  printf("Wrist     base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_wrist[0],c_wrist[1],c_wrist[2]);
  printf("Lower arm base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_lowerarm[0],c_lowerarm[1],c_lowerarm[2]);
  printf("Upper arm base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_upperarm[0],c_upperarm[1],c_upperarm[2]);
  printf("CCD Angles\n[S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E2=%5.1f] [W1=%5.1f W2=%5.1f]\n",
      ccd_rs1,ccd_rs2,ccd_rs3,ccd_re1,ccd_re2,ccd_rw1,ccd_rw2);
  printf("CCD EndPoint Transform\n");
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[0][0],BULH[0][1],BULH[0][2],BULH[0][3]); 
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[1][0],BULH[1][1],BULH[1][2],BULH[1][3]); 
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[2][0],BULH[2][1],BULH[2][2],BULH[2][3]); 
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[3][0],BULH[3][1],BULH[3][2],BULH[3][3]); 

  {float h[4],X[4]={1,0,0,1},Y[4]={0,1,0,1},Z[4]={0,0,1,1};
   ExtractMatrixRotation(BULH,mmr);   
   m4by1(mmr,X,h);  printf("Hand Axes     : X=(%6.2f %6.2f %6.2f)\n",h[0],h[1],h[2]);
   m4by1(mmr,Y,h);  printf("              : Y=(%6.2f %6.2f %6.2f)\n",h[0],h[1],h[2]);
   m4by1(mmr,Z,h);  printf("              : Z=(%6.2f %6.2f %6.2f)\n",h[0],h[1],h[2]);
  }
 }
}

///////////////  Get Information /////////////////////////////////

void GetRobotTransform(GLfloat me[4][4], GLfloat mp[4][4], GLfloat mua[4][4], GLfloat mla[4][4]){
// Retrieve the OpenGL transformation matrices of the parts of the Robot arm and convert them
// into C order 4 x 4 matrices
//
//     0 1 2 3 4 5 ....  Linear (vector) form
//
//     0   1   2   3     [0][0]  [0][1]  [0][2]  [0][3]  
//     4   5   6   7     [1][0]  [1][1]  [1][2]  [1][3]
//     8   9  10  11     [2][0]  [2][1]
//    12  13  14  15     [3][0]  [3][1]  

//  OGL ordering is TRANSPOSED when copied from vector (linear) form
 int i;
 GLfloat cameraViewMatrix[16],m[16],pe[4],zero[4]={0.0,0.0,0.0,1.0};
 glGetFloatv(GL_MODELVIEW_MATRIX, cameraViewMatrix);
 glLoadIdentity();

 glRotatef(rs1,0.0,0.0,1.0);  // first axis  S1  around local z  (applies to all children)
 glTranslatef(0.0f,0.0f,baseHeight);
 glRotatef(rs2,0.0,1.0,0.0);   // second axis S2  around local y
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   
 m16to4(m,mua);  // pivot point (upperarm)

 glRotatef(rs3,0.0,0.0,1.0);   // 3rd axis S3
 glTranslatef(0.0f,0.0f,upperArm);
 glRotatef(re1,0.0,1.0,0.0);  // 4th axis E1
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   
 m16to4(m,mla);  // pivot point (lowerarm)

 glRotatef(re2,0.0,0.0,1.0);   // 5th axis E2
 glTranslatef(0.0f,0.0f,lowerArm);
 glRotatef(rw1,0.0,1.0,0.0);  // 6th axis W1
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   // to go from OGL order to C++ ordering
 m16to4(m,mp);  // pivot point (wrist)

 glRotatef(rw2,0.0,0.0,1.0);   // 7th axis W2
 glTranslatef(0.0f,0.0f,Hand);
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   // to go from OGL order to C++ ordering
 m16to4(m,me);

 glMatrixMode(GL_MODELVIEW);
 glLoadMatrixf(cameraViewMatrix);
 get_transform=0;  // reset the flag

 printf("Get Open GL Transform\n");
 printf("Angles:\n[S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E1=%5.1f] [W1=%5.1f W2=%5.1f]\n",
        rs1,rs2,rs3,re1,re2,rw1,rw2);
 printf("Transform Matrix End Point\n");
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[0][0],me[0][1],me[0][2],me[0][3]); 
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[1][0],me[1][1],me[1][2],me[1][3]); 
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[2][0],me[2][1],me[2][2],me[2][3]); 
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[3][0],me[3][1],me[3][2],me[3][3]);

 m4by1(me,zero,pe);
 printf("Graphic End point:(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
 m4by1(mp,zero,pe);
 p_wrist[0]=pe[0]; p_wrist[1]=pe[1]; p_wrist[2]=pe[2];
 printf("Graphic Wrist    :(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
 m4by1(mla,zero,pe);
 p_lowerarm[0]=pe[0]; p_lowerarm[1]=pe[1]; p_lowerarm[2]=pe[2];
 printf("Graphic Lower arm:(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
 m4by1(mua,zero,pe);
 p_upperarm[0]=pe[0]; p_upperarm[1]=pe[1]; p_upperarm[2]=pe[2];
 printf("Graphic Upper arm:(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
}

//////////////////////////////  DRAWING functons  Draw roobt and Axes ///

void drawDisk(float r) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  gluDisk(pObj,0.0f,r,30,3);
  gluDeleteQuadric(pObj);
  glPopMatrix();
}

void drawSphere(float r) {
  GLUquadricObj* pObj = gluNewQuadric();
  gluSphere(pObj, r, 30, 30);
  gluDeleteQuadric(pObj);
}

void drawCylinder(float r, float h) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  gluCylinder(pObj,r,r,h,30,10);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(0.0f,0.0f,h);
  gluDisk(pObj,0.0f,r,30,3);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,30,3);
  glPopMatrix();
  gluDeleteQuadric(pObj);
}

void drawWireCylinder(float r, float h, float b) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  gluQuadricDrawStyle(pObj,GLU_LINE);
  //  gluQuadricDrawStyle(pObj,GLU_SILHOUETTE);
  glLineWidth(0.25);
  glTranslatef(0.0f,0.0f,b);
  gluCylinder(pObj,r,r,h,15,5);
  glTranslatef(0.0f,0.0f,h);
  gluDisk(pObj,0.0f,r,15,1);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,15,1);
  glPopMatrix();
  gluDeleteQuadric(pObj);
}

void drawCylinderOffset(float r, float h, float b) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  glColorMaterial(GL_FRONT,GL_DIFFUSE);  
  glEnable(GL_COLOR_MATERIAL);
  glColor3f(1.0,0.8,0.8);
  glTranslatef(0.0f,0.0f,b);
  gluCylinder(pObj,r,r,h,15,5);
  glTranslatef(0.0f,0.0f,h);
  gluDisk(pObj,0.0f,r,15,1);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,15,1);
  glPopMatrix();
  glColor3f(1.0,1.0,1.0);
  glDisable(GL_COLOR_MATERIAL);
  gluDeleteQuadric(pObj);
}

void drawCylinderH(float r, float h) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  glRotatef(90.0,1.0,0.0,0.0);
  glPushMatrix();
  glTranslatef(0.0f,0.0f,-h/2.0);
  gluCylinder(pObj,r,r,h,30,10);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(0.0f,0.0f,h/2.0);
  gluDisk(pObj,0.0f,r,30,3);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,30,3);
  glPopMatrix();
  glPopMatrix();
  gluDeleteQuadric(pObj);
}

void drawAxes(float l){  // draw an rectangular axis in the current frame of reference X=RED Y=GREEEN Z=BLUE
   glDisable(GL_LIGHTING);
   glColorMaterial(GL_FRONT,GL_DIFFUSE);  
   glEnable(GL_COLOR_MATERIAL);
   glLineWidth(4.0);
   glColor3f(1.0,0.0,0.0);
   glBegin(GL_LINES);
   glVertex3f(0.0,0.0,0.0);
   glVertex3f(l,  0.0,0.0);
   glColor3f(0.0,1.0,0.0);
   glBegin(GL_LINES);
   glVertex3f(0.0,0.0,0.0);
   glVertex3f(0.0,  l,0.0);
   glColor3f(0.0,0.0,1.0);
   glBegin(GL_LINES);
   glVertex3f(0.0,0.0,0.0);
   glVertex3f(0.0,0.0, l);
   glEnd();
   glColor3f(1.0,1.0,1.0);
   glDisable(GL_COLOR_MATERIAL);
   glEnable(GL_LIGHTING);
}

void drawTarget(void){
  float mm[4][4],mmm[16],mmi[4][4],mmr[4][4],mmu[4][4];
#if 1
  glPushMatrix();     // draw AXIS at the tool position
  glLoadIdentity(); 
  //
  //ExtractMatrixRotation(m_endpoint,mmr);  // These steps undo the rotation part of the end
  //InvertMatrixRotation(mmr,mmi,1);        // effector so that it aligns with global 
  //m4by4(mmi,mmr,mmu);                     // frame of reference (at the base of the robot.)
  //SetMatrixRotation(m_endpoint,mmu,mmr);  // ..
  //m4by4(m_camera,mmr,mm);                 // .. They could replace the next line
  m4by4(m_camera,m_endpoint,mm);  // compose position and camera view transform
  //
  m4to16(mm,mmm);
  TransposeMatrix16(mmm);
  glLoadMatrixf(mmm);
  drawAxes(15.0);
  glPopMatrix();      // end of endpoint axes
#endif

  glColorMaterial(GL_FRONT,GL_DIFFUSE);  
  glEnable(GL_COLOR_MATERIAL); // draw the target
  glColor3f(1.0,1.0,0.5);
  glPushMatrix();     // target for CCD algorithm to place wrist joint at 
  glTranslatef(p_target[0],p_target[1],p_target[2]);
  drawSphere(1.2);
  glPopMatrix();    
  glColor3f(1.0,1.0,1.0);
  glDisable(GL_COLOR_MATERIAL);
}

void drawPivot(float disp, float height, float width, float offset){
 drawCylinderOffset(disp,1.0,offset);
 glPushMatrix();
 glTranslatef(0.0,-disp,offset+height/2.0);
 glScalef(width,2.0,height);
 glutSolidCube(1.0);
 glPopMatrix();
 glPushMatrix();
 glTranslatef(0.0, disp,offset+height/2.0);
 glScalef(width,2.0,height);
 glutSolidCube(1.0);
 glPopMatrix();
}

void drawCCDAxes() {
 GLfloat m[16],m4[4][4];
 drawAxes(30.0);   // base axis 
 drawSphere(2.0); 
 glPushMatrix();
 m4by4(m_camera,B,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 m4by4(m_camera,BU,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 m4by4(m_camera,BUL,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 m4by4(m_camera,BULH,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 glPopMatrix();
}

void drawRobotAxes() {
 drawAxes(30.0);   // base axis   [I]
 glPushMatrix();

 glRotatef(rs1,0.0,0.0,1.0);  // first axis  S1  around local z  (applies to all children)
 glTranslatef(0.0f,0.0f,baseHeight);
 glRotatef(rs2,0.0,1.0,0.0);   // second axis S2  around local y
 drawAxes(15.0);   // first joint

 glRotatef(rs3,0.0,0.0,1.0);   // 3rd axis S3
 glTranslatef(0.0f,0.0f,upperArm);
 glRotatef(re1,0.0,1.0,0.0);  // 4th axis E1
 drawAxes(10.0);  

 glRotatef(re2,0.0,0.0,1.0);   // 5th axis E2
 glTranslatef(0.0f,0.0f,lowerArm);
 glRotatef(rw1,0.0,1.0,0.0);  // 6th axis W1
 drawAxes(10.0);

 glRotatef(rw2,0.0,0.0,1.0);   // 7th axis W2

 glTranslatef(0.0f,0.0f,Hand); 
 drawAxes(5.0);   // Final axis  (M)
 drawSphere(1.0); 

 glPopMatrix();
}

void drawRobotGeometry(void) {
 float Lrs1,Lrs2,Lrs3,Lre1,Lre2,Lrw1,Lrw2;
 Lrs1=rs1; Lrs2=rs2; Lrs3=rs3; Lre1=re1; Lre2=re2; Lrw1=rw1; Lrw2=rw2;

 glPushMatrix();      // draw base disk
 glTranslatef(0.0,0.0,-0.5);
 drawDisk(75.0);
 glPopMatrix();       // end of base disk

 drawCylinder(15.0,15.0);
 glPushMatrix();
 glRotatef(Lrs1,0.0,0.0,1.0);  // first axis  S1  around local z  (applies to all children)
 drawPivot(8,baseHeight-15.0+5.0,10.0, 15.0);
 glTranslatef(0.0f,0.0f,baseHeight);
 drawCylinderH(5.0,15.0);
 drawCylinderH(1.0,20.0);
 glRotatef(Lrs2,0.0,1.0,0.0);   // second axis S2  around local y
 if(axesVis)drawWireCylinder(5,upperArm-15.0,0.0); else drawCylinder(5,upperArm-15.0);  

 glRotatef(Lrs3,0.0,0.0,1.0);   // 3rd axis S3
 drawPivot(8,15.0,10.0, upperArm-15.0);

 glTranslatef(0.0f,0.0f,upperArm);
 drawCylinderH(3.0,12.0);
 drawCylinderH(1.0,15.0);
 glRotatef(Lre1,0.0,1.0,0.0);  // 4th axis E1
 if(axesVis)drawWireCylinder(5,lowerArm-20.0,0.0);  else drawCylinder(5,lowerArm-20.0);

 glRotatef(Lre2,0.0,0.0,1.0);   // 5th axis E2
 drawPivot(6,20.0,6.0, lowerArm-20.0);
 
 glTranslatef(0.0f,0.0f,lowerArm);
 drawCylinderH(1.0,18.0);  // pin
 glRotatef(Lrw1,0.0,1.0,0.0);  // 6th axis W1
 drawCylinderOffset(2.5,12.0,-12.0);

 glRotatef(Lrw2,0.0,0.0,1.0);   // 7th axis W2
 if(axesVis)drawWireCylinder(3.0,8.0,0.0); else drawCylinder(3.0,8.0);  //The hand
 
 glTranslatef(0.0f,0.0f,Hand);
 if(!axesVis){
   drawSphere(2.0);
   drawWireCylinder(1.0,3.0,0.0);
 }
 glPopMatrix();
}

static void display(void) {   // this is the GLUT idle callback function, it is used to draw the robot
  static GLfloat light0_ambient[] =  {0.1f, 0.1f, 0.3f, 1.0f};
  static GLfloat light0_diffuse[] =  {.8f, .8f, 1.0f, 1.0f};
  static GLfloat light0_position[] = {.5f, .5f, 1.0f, 0.0f};
  GLfloat m[16];
  GLdouble ex,ey,ez;
	glClearDepth(1.0f);
  //	glClearColor(0.1f, 0.1f, 0.2f, 0.0f);
	glClearColor(1.0f, 1.0f, 1.0f, 0.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
	glShadeModel(GL_SMOOTH);
  glEnable( GL_LIGHTING );
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse );
  glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient );

  glPushMatrix();
  glLoadIdentity();
  if(viewdir == 0){ // set the appropriate view direction transformation
    ez=er*cos(theta)*cos(phi);
    ex=er*cos(theta)*sin(phi);
    ey=er*sin(theta);
    gluLookAt(ex, ey, ez+50, 0, 0, 50, 0, 1, 0);    // y - up
  }
  else if(viewdir == 1){
    ex=er*cos(theta)*cos(phi);
    ey=er*cos(theta)*sin(phi);
    ez=er*sin(theta);
    gluLookAt(ex, ey, ez+50, 0, 0, 50, 0, 0, 1);    // z - up
  }
  else if(viewdir == 2){
    ey=er*cos(theta)*cos(phi);
    ez=er*cos(theta)*sin(phi);
    ex=er*sin(theta);
    gluLookAt(ex, ey, ez+50, 0, 0, 50, 1, 0, 0);    // x - up
  }
  glEnable( GL_NORMALIZE );

  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);  // light stays fixed relative to objects 

  // Just to make sure our transform functions are OK,
  // read the OGL transformation convert and convert back.
  glGetFloatv(GL_MODELVIEW_MATRIX, m);
  TransposeMatrix16(m);  // OGL to C matrix order  
  m16to4(m,m_camera);    // model view matrix of camera
  m4to16(m_camera,m);
  TransposeMatrix16(m);  // C to OGL matrix ordering  
  glMatrixMode(GL_MODELVIEW);
  glLoadMatrixf(m);
  ///////////////// end of digression
  drawTarget(); // draw the IK target point

  if(modelVis){// if the model is visible 
    drawRobotGeometry();
  }
  if(axesVis)drawRobotAxes(); // if the axes are to be drawn
  if(ccdVis)drawCCDAxes();  // show the axes generated by the IK algorithm (onlhy valid when IK is in operation)
  // read back the OpenGL transformations (used for comparision purposes)
  if(get_transform)GetRobotTransform(m_endpoint,m_wristpoint,m_upperarm,m_lowerarm);
  glPopMatrix();
  glutSwapBuffers();
}

void reshape(GLint width, GLint height){ // GLUT function called when the window is re-sized
   glViewport(0, 0, width, height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(65.0, (float)width / height, 3, 800);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

// User interface functions Mouse and Keyboard /////////////////////////////////////

void Special(int key, int x, int y){  // special keyboard keys
  int bchange=0,bShift=0;
  float step_size=0.1;
  float angl_step=0.5;
  unsigned long bmModifiers;
  bmModifiers = glutGetModifiers();
  if((bmModifiers & GLUT_ACTIVE_SHIFT) == GLUT_ACTIVE_SHIFT)bShift=1; else bShift=0;
  if((bmModifiers & GLUT_ACTIVE_CTRL) == GLUT_ACTIVE_CTRL)step_size=0.15; else step_size=1.5;
  if(key == GLUT_KEY_RIGHT){  // arrow keys
     if(bShift)hand_pan += angl_step;     // change the direction of the tool
     else      p_target[1] += step_size;  // move the IK target point
     bchange=1;
  } 
  if(key == GLUT_KEY_LEFT) {
     if(bShift)hand_pan -= angl_step;
     else     p_target[1] -= step_size;  
     bchange=1;
  }
  if(key == GLUT_KEY_UP)   {
     if(bShift)hand_tilt += angl_step;
     else     p_target[0] += step_size;  
     bchange=1;
  }
  if(key == GLUT_KEY_DOWN) {
     if(bShift)hand_tilt -= angl_step;
     else     p_target[0] -= step_size;  
     bchange=1;
  }  
  if(key == GLUT_KEY_PAGE_UP){
     if(bShift)hand_twist += angl_step;
     else     p_target[2] += step_size;  
     bchange=1;
  }
  if(key == GLUT_KEY_PAGE_DOWN){
     if(bShift)hand_twist -= angl_step;
     else     p_target[2] -= step_size;  
     bchange=1;
  }
  if(bchange && interactive){
    RunCCD(0);
  }
}

void Keyboard(unsigned char key, int x, int y){ // normal keyboard keys
  if(key==27) { // ESC
    exit(0);
  }
  // change the pivot angles using these keys
  if(key=='1')rs1 += 2.0;  
  if(key=='2')rs1 -= 2.0;
  if(key=='3')rs2 += 2.0;
  if(key=='4')rs2 -= 2.0;
  if(key=='5')rs3 += 2.0;
  if(key=='6')rs3 -= 2.0;
  if(key=='7')re1 += 2.0;
  if(key=='8')re1 -= 2.0;
  // change the tool (wrist) angles using these keys
  if(key=='9')re2 += 2.0;  // twist wrist
  if(key=='0')re2 -= 2.0;
  if(key=='-')rw1 += 2.0;  // hand up/down
  if(key=='=')rw1 -= 2.0;
  if(key=='q')rw2 += 2.0;  // hand twist angle 
  if(key=='w')rw2 -= 2.0;
  // change viewing orientation
  if(key=='y')viewdir=0;
  if(key=='z')viewdir=1;
  if(key=='x')viewdir=2;
  if(key=='~'){
    printf("Angles: [S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E2=%5.1f] [W1=%5.1f W2=%5.1f]\n",
     rs1,rs2,rs3,re1,re2,rw1,rw2);
  }
  if(key=='.'){     // set the flag to read back the OpenGL gransformations
    get_transform=1;
  }
  if(key=='@'){    // runs the IK algorithm - displaying individual steps
    RunCCD(1);
  }
}

static int bMouseDown=0,bMouseRightDown, bShiftModifier=0;
static int xd,yd,xrd,yrd;

void mouse(int button, int state, int x, int y){
   //#define GLUT_ACTIVE_CTRL                2
   //#define GLUT_ACTIVE_ALT                 4
   unsigned long bmModifiers;
   bmModifiers = glutGetModifiers();
   if((bmModifiers & GLUT_ACTIVE_SHIFT) == GLUT_ACTIVE_SHIFT)bShiftModifier=1; else bShiftModifier=0;
   if (button == GLUT_LEFT_BUTTON)   {
      if (state == GLUT_UP)      {
        bMouseDown=0;
      }
      else {
        xd=x; yd=y;
        bMouseDown=1;
      }
   }
   if (button == GLUT_RIGHT_BUTTON)   {
      if (state == GLUT_UP)      {
        bMouseRightDown=0;
      }
      else {
        xrd=x; yrd=y; 
        bMouseRightDown=1;
      }
   }
   if (button == GLUT_MIDDLE_BUTTON)   {
      if (state == GLUT_UP)      {
      }
      else {
      }
   }
}

void motion(int x, int y){
  GLdouble dx,dy,dr;
  if(bMouseDown == 1){
    dx=x-xd; dy=y-yd;
    if(bShiftModifier){
      phi   -= dx*0.03;
      theta += dy*0.04;
    }
    else{
      phi   -= dx*0.003;
      theta += dy*0.004;
    }
    xd=x; yd=y;
  }
  if(bMouseRightDown == 1){
    dr=x-xrd;
    er -= dr*0.05;
    xrd=x;
  }
}

void menuCallback(int value){
  switch (value) {
    case 100:
      modelVis ^= 1;
      glutPostRedisplay();
      break;
    case 104:
      axesVis ^= 1;
      glutPostRedisplay();
      break;
    case 114:
      ccdVis ^= 1;
      glutPostRedisplay();
      break;
    case 101:
      viewdir=0;
      glutPostRedisplay();
      break;
    case 102:
      viewdir=1;
      glutPostRedisplay();
      break;
    case 103:
      viewdir=2;
      glutPostRedisplay();
      break;
    case 109:
      printf("Angles: [S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E1=%5.1f] [W1=%5.1f W2=%5.1f]\n",
            rs1,rs2,rs3,re1,re2,rw1,rw2);
      break;
    case 111:
      get_transform=1;
      break;
    case 110:
      rs1=0.0,rs2=0.0,rs3=0.0,re1=0.0,re2=0.0,rw1=0.0,rw2=0.0;
      hand_pan=hand_tilt=0.0; hand_twist=0.0;
      glutPostRedisplay();
      break;
    case 115:
      RunCCD(1);
      break;
    case 116:
      interactive ^=1;
      break; 
    case 666:
      exit(0);
      break;
    default:
      break;
  }
}

static void AddMenuItems(void){
  int shapeMenu, viewMenu, actionMenu, renderModeMenu;
  shapeMenu = glutCreateMenu(menuCallback);
  glutAddMenuEntry("Model", 100);
  glutAddMenuEntry("Axes", 104);
  glutAddMenuEntry("CCD Axes", 114);
  viewMenu = glutCreateMenu(menuCallback);
  glutAddMenuEntry("Y - up ", 101);
  glutAddMenuEntry("Z - up ", 102);
  glutAddMenuEntry("X - up ", 103);
  renderModeMenu = glutCreateMenu(menuCallback);
  actionMenu = glutCreateMenu(menuCallback);
  glutAddMenuEntry("Position at Target (@)",115);
  glutAddMenuEntry("Print Angles (#)", 109);
  glutAddMenuEntry("Print Matrices / Move End Proxy (.)", 111);
  // create the context menu
  glutCreateMenu(menuCallback);
  glutAddSubMenu("Hide/Show", shapeMenu);
  glutAddSubMenu("View Direction  ", viewMenu);
  glutAddSubMenu("Action", actionMenu);
  glutAddMenuEntry("Reset",110);
  glutAddMenuEntry("Toggle Interactive Target Tracking",116);
  glutAddMenuEntry("Quit", 666);
  // attach the main menu to the Middle mouse button
  glutAttachMenu(GLUT_MIDDLE_BUTTON);
}

///////////  end UI stuff


//////////////////  Geometry & Utility Functions belowe //////////
//////////////////////////////////////////////////////////////////

static void RotateVector(float ct, float st, float r[4], float n[4]){
 int i;
 float t1[4],t2[4],t3[4],t4[4],cc,ndr;
 VECSCALE4(ct,r,t1)
 ndr=DOT4(n,r);
 cc=(1.0-ct)*ndr;
 VECSCALE4(cc,n,t2)
 CROSS4(n,r,t3)
 VECSCALE(st,t3,t4)
 for(i=0;i<4;i++){
   r[i]=t1[i]+t2[i]+t4[i];
 }
}

// use of this SHOULD BE REPLACE BY funciton that obtains the rotation inverse using
// THE TRANSPOSE since it is the SAME as the full inverse for ROTATION (i.e. symmetric matrix)
static void InvertMatrixRotation(float T[4][4], float t[4][4], int zero_translation){
 int i,j;
 float a[3][3];
 float det;
 null_transform(t);
 if(zero_translation){for(i=0;i<3;i++)for(j=0;j<3;j++)a[i][j]=T[i][j];}
 else                {for(i=0;i<4;i++)for(j=0;j<4;j++)a[i][j]=T[i][j];}
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
}

static void InvertMatrixRotationT(float T[4][4], float t[4][4], int zero_rotation){ // using transpose
 int i,j;
 null_transform(t);
 if(!zero_rotation){for(i=0;i<4;i++){t[i][3]=T[i][3]; t[3][i]=T[3][i];  } }
 for(i=0;i<3;i++) for(j=0;j<3;j++) {
   t[i][j]=T[j][i];
 }
}

static void ExtractMatrixRotation(float T[4][4], float t[4][4]){  // get the rotation part of a matrix
 int i,j;
 null_transform(t);
 for(i=0;i<3;i++)for(j=0;j<3;j++)t[i][j]=T[i][j];
}
 
static void SetMatrixRotation(float T[4][4], float r[4][4], float t[4][4]){ // set the rotation part of matrix
 int i,j;
 for(i=0;i<4;i++)for(j=0;j<4;j++)t[i][j]=T[i][j];
 for(i=0;i<3;i++)for(j=0;j<3;j++)t[i][j]=r[i][j];
}


static void TransposeMatrix16(GLfloat m[16]){
  GLfloat t;
  t=m[ 4]; m[ 4]=m[ 1]; m[ 1]=t;
  t=m[ 6]; m[ 6]=m[ 9]; m[ 9]=t;
  t=m[11]; m[11]=m[14]; m[14]=t;
  t=m[ 2]; m[ 2]=m[ 8]; m[ 8]=t;
  t=m[ 3]; m[ 3]=m[12]; m[12]=t;
  t=m[ 7]; m[ 7]=m[13]; m[13]=t;
}

static void m4by4(GLfloat t1[4][4], GLfloat t2[4][4], GLfloat tr[4][4]){   //tr = t1 * t2
  // tr = t1 X t2
 long i,j,k;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;

  for(k=0;k<4;k++)tr[i][j]=tr[i][j]+t1[i][k]*t2[k][j];
 }
 return;
}

static void m4by1(GLfloat t4[4][4], GLfloat m[4],  GLfloat r[4]){  // C++ ordering
   r[0]=t4[0][0]*m[0]+t4[0][1]*m[1]+t4[0][2]*m[2]+t4[0][3]*m[3];
   r[1]=t4[1][0]*m[0]+t4[1][1]*m[1]+t4[1][2]*m[2]+t4[1][3]*m[3];
   r[2]=t4[2][0]*m[0]+t4[2][1]*m[1]+t4[2][2]*m[2]+t4[2][3]*m[3];
   r[3]=t4[3][0]*m[0]+t4[3][1]*m[1]+t4[3][2]*m[2]+t4[3][3]*m[3];
}

static void m16to4(GLfloat m[16], GLfloat v[4][4]){  // C++ ordering
  v[0][0]=m[ 0]; v[0][1]=m[ 1]; v[0][2]=m[ 2]; v[0][3]=m[ 3];
  v[1][0]=m[ 4]; v[1][1]=m[ 5]; v[1][2]=m[ 6]; v[1][3]=m[ 7];
  v[2][0]=m[ 8]; v[2][1]=m[ 9]; v[2][2]=m[10]; v[2][3]=m[11];
  v[3][0]=m[12]; v[3][1]=m[13]; v[3][2]=m[14]; v[3][3]=m[15];
}

static void m4to16(GLfloat v[4][4], GLfloat m[16]){   // C++ ordering
  m[ 0]= v[0][0]; m[ 1]=v[0][1]; m[ 2]=v[0][2]; m[ 3]=v[0][3];
  m[ 4]= v[1][0]; m[ 5]=v[1][1]; m[ 6]=v[1][2]; m[ 7]=v[1][3];
  m[ 8]= v[2][0]; m[ 9]=v[2][1]; m[10]=v[2][2]; m[11]=v[2][3];
  m[12]= v[3][0]; m[13]=v[3][1]; m[14]=v[3][2]; m[15]=v[3][3];
}

static void rotz(float tr[4][4], float ang){
 long i,j;
 ang *= DEGRAD;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][1] = -sin(ang);
  tr[1][0] =  sin(ang);
  tr[1][1] =  cos(ang);
  return;
}

static void rotx(float tr[4][4], float ang){
 long i,j;
 ang *= DEGRAD;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[1][1] =  cos(ang);
  tr[1][2] = -sin(ang);
  tr[2][1] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

static void roty(float tr[4][4], float ang){
 long i,j;
 ang *= DEGRAD * (-1.0);   
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][2] = -sin(ang);
  tr[2][0] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

static void tram(float t[4][4], float dx, float dy, float dz){
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  t[i][j]=0.0;
  if(i == j)t[i][j]=1.0;
 }
 t[0][3]=dx;
 t[1][3]=dy;
 t[2][3]=dz;
 t[3][3]=1;
 return;
}

static void c4to4(float tin[4][4], float tout[4][4]){
 int i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 tout[i][j]=tin[i][j];
 return;
}

static void null_transform(float t[4][4]){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 if(i == j)t[i][j]=1.0;
 else      t[i][j]=0.0;
}


static void normalise(float v[4]){
 float n,nn;
 n= (float)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-15)return;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE4(n,v,v)
 return;
}

////////////// // entry point  (configure the GLUT

int main(int argc, char* argv[]) {
  int window_id;
  glutInit(&argc, argv);
  glutInitWindowPosition(400,300);
  glutInitWindowSize(800,600);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  window_id = glutCreateWindow( "Robot Simulator" );
  glutSetWindow( window_id );
  glutDisplayFunc ( display  );  // draw the window
  glutReshapeFunc (reshape);
  // keyboard functions 
  glutKeyboardFunc(Keyboard);
  glutSpecialFunc(Special);
  glutIdleFunc( display );   // continue to update - called because of interactive IK 
  // mouse functions
  glutMotionFunc(motion);
  glutMouseFunc(mouse);
  // context menu
  AddMenuItems();
  glutMainLoop();
  printf("Program Ended\n");
}
